//usage:
// * define Game.prototype.init() and Game.prototype.update(dt)
// * game = new Game(includes), where includes is a list of scripts to include

Math.TAU = Math.PI * 2;

function Game(includes){
	//graphics
	this.width = 800; this.height = 600;
	this.canvas = document.getElementById("game-window");
	this.canvas.width = this.width; this.canvas.height = this.height;
	this.ctx = this.canvas.getContext("2d");
	this.dt_max = 1/30;
	this.anim_ref = null;
	
	//other scripts
	if (includes !== undefined){
		this.progress = includes.length+1;
		for (var i = 0; i < includes.length; i++){
			this.load_script(includes[i], this.run.bind(this));
		}
	}else{
		this.progress = 1;
	}
	
	this.update_cb = this._update.bind(this);
	this.init();
	this.run();
}

Game.prototype.load_script = function(src, callback){
	var script = document.createElement('script');
	if (src.substr(-3) !== ".js")
		src += ".js";
	script.src = src;
	
	if (script.readyState){ // IE, incl. IE9
		script.onreadystatechange = function(){
			if (script.readyState == "loaded" || script.readyState == "complete"){
				script.onreadystatechange = null;
				callback();
			}
		}
	}else{ // Other browsers
		script.onload = function(){
			callback();
		}
	}
	document.body.appendChild(script);
}

Game.prototype.run = function(){
	if (--this.progress <= 0){
		if (this.anim_ref === null)
			this.anim_ref = window.requestAnimationFrame(this.update_cb);
	}
}

Game.prototype._update = function(now){
	if (!this.now)
		this.now = now;
	var dt = Math.min((now - this.now)/1000, this.dt_max);
	this.now = now;
	
	if (this.update(dt))	//update should return true for 'done'
		this.anim_ref = null;
	else
		this.anim_ref = window.requestAnimationFrame(this.update_cb);
}
