Array.prototype.eradicate = function(e){	//remove all occurences of e from array
	var i;
	while (true){
		i = this.indexOf(e);
		if (i >= 0)
			this.splice(i, 1);
		else
			break;
	}
}

function getJson(url, callback){
	var req = new XMLHttpRequest();
	req.open("get", url, true);
	req.responseType = "json";
	req.onload = function(){callback(req.response)};
	req.send();
}

//-------- extra objects

//sprite

function Sprite(img, x, y, scale, n_frames, df, shadow){
	// this.img = img;
	this.img_name = img;
	this.x = x;
	this.y = y;
	this.w_f = this.img.width;	//frame dimensions
	this.h_f = this.img.height;	//
	this.nf = n_frames || 1;
	if (this.nf > 1){
		this.w_f = Math.floor(this.w_f/this.nf);
		this.f = 0;
		this.df = df || 1;	//rate of change of f
	}
	this.scale = scale || 1;
	this.w = this.w_f * this.scale;
	this.h = this.h_f * this.scale;
	
	if (shadow){
		this.shadow = new Sprite("shadow", 0, 0, scale);
		this.shadow.x_off = (this.w - this.shadow.w)/2;
		this.shadow.y_off = this.h - this.shadow.h;
	}
}

Object.defineProperties(Sprite.prototype, {
	"img_name": {get: function(){return this._img_name}, set: function(img){if (this._img_name !== img){this._img_name = img; this.img = this._imgs[img];}}},
	"left": {get: function(){return this.x}, set: function(x){this.x = x}},
	"top": {get: function(){return this.y}, set: function(y){this.y = y}},
	"right": {get: function(){return this.x + this.w}, set: function(x){this.x = x - this.w}},
	"bottom": {get: function(){return this.y + this.h}, set: function(y){this.y = y - this.h}},
	"mid_x": {get: function(){return this.x + this.w/2}, set: function(x){this.x = x - this.w/2}},
	"mid_y": {get: function(){return this.y + this.h/2}, set: function(y){this.y = y - this.h/2}}
});

Sprite.prototype.animate = function(dt){
	if (this.nf > 1){
		var range = this.state_frames[this.state];
		this.f += this.df * dt;
		if (this.f < range[0])
			this.f = range[0];
		else if (this.f > range[1])
			this.f = range[0] + this.f%1;
	}
}

Sprite.prototype.draw = function(ctx){
	if (this.shadow){
		this.shadow.x = this.x+this.shadow.x_off;
		this.shadow.y = this.y+this.shadow.y_off;
		this.shadow.draw(ctx);
	}
	if (this.nf > 1)
		ctx.drawImage(this.img, Math.floor(this.f)*this.w_f, 0, this.w_f, this.h_f, Math.floor(this.x), Math.floor(this.y), this.w, this.h);
	else
		ctx.drawImage(this.img, Math.floor(this.x), Math.floor(this.y), this.w, this.h);
}

//room

function Room(contents, grid_width, grid_height, scale){
	floor_img = this._imgs["floor"+contents["floor"]];
	wall_img = this._imgs["wall"+contents["wall"]];
	this.canvas = document.createElement("canvas");
	var tile_size = floor_img.width;
	grid_width += 2;
	grid_height += 2;
	this.canvas.width = grid_width*tile_size;
	this.canvas.height = grid_height*tile_size;
	this.grid_width = grid_width;
	this.grid_height = grid_height;
	this.full_width = this.canvas.width*scale;
	this.full_height = this.canvas.height*scale;
	this.tile_size = tile_size*scale;
	this.x = 0;
	this.y = 0;
	
	var ctx = this.canvas.getContext("2d"), x, y, sx, sy;
	this.occupied = new Array(grid_width);
	this.interaction = new Array(grid_width);
	for (x = 0; x < grid_width; x++){
		this.occupied[x] = new Array(grid_height);
		this.interaction[x] = new Array(grid_height);
		for (y = 0; y < grid_height; y++){
			sx = x > 0 ? x < grid_width-1 ? 1 : 2 : 0;
			sy = y > 0 ? y < grid_height-1 ? 2 : 3 : 1;
			if (sx === 1 && sy === 2){
				ctx.drawImage(floor_img, x*tile_size, y*tile_size);
				this.occupied[x][y] = false;
			}else{
				ctx.drawImage(wall_img, sx*tile_size, sy*tile_size, tile_size, tile_size, x*tile_size, y*tile_size, tile_size, tile_size);
				this.occupied[x][y] = true;
			}
		}
	}
	
	this.sprites = [];
	var i, thing, sprite, action;
	if (contents.hasOwnProperty("tele")){
		for (i = 0; i < contents["tele"].length; i++){
			thing = contents["tele"][i];
			sprite = new Sprite("teleporter", 0, 0, scale, 2, 0);
			sprite.grid_x = thing.x;
			sprite.grid_y = thing.y;
			sprite.x_off = thing.x*this.tile_size;
			sprite.y_off = (thing.y+1)*this.tile_size - sprite.h;
			sprite.state = thing.state;
			sprite.state_frames = {"off": [0, 1], "on": [1, 2]};
			sprite.f = sprite.state === "on" ? 1 : 0;
			this.interaction[thing.x][thing.y] = {
				"type": sprite.state === "on" ? "teleport" : "teleport-inactive",
				"target": thing.target,
				"sprite": sprite	//just because it needs to be referenced when activating
			}
			this.sprites.push(sprite);
		}
	}
	if (contents.hasOwnProperty("npc")){
		for (i = 0; i < contents["npc"].length; i++){
			thing = contents["npc"][i];
			sprite = new Sprite(thing.type, 0, 0, scale, thing.frames, 0, true);
			this.occupied[thing.x][thing.y] = true;
			sprite.grid_x = thing.x;
			sprite.grid_y = thing.y;
			sprite.x_off = thing.x*this.tile_size;
			sprite.y_off = (thing.y+1)*this.tile_size - sprite.h;
			sprite.dialogue = thing.dialogue;
			sprite.wants = thing.wants;
			sprite.response = thing.response;
			sprite.reaction = thing.reaction;
			this.interaction[thing.interaction.x][thing.interaction.y] = {
				"type": "speak",
				"target": sprite
			}
			this.sprites.push(sprite);
		}
	}
}

Room.prototype.draw = function(ctx){
	ctx.drawImage(this.canvas, Math.floor(this.x), Math.floor(this.y), this.full_width, this.full_height);
	var i, y, sprite;
	for (y = 1; y < this.grid_height-1; y++){
		for (i = 0; i < this.sprites.length; i++){
			sprite = this.sprites[i];
			if (sprite.grid_y === y){
				if (sprite.hasOwnProperty("x_off")){	//bit of a hacky way of excluding the player but oh well...
					sprite.x = this.x+sprite.x_off;
					sprite.y = this.y+sprite.y_off;
				}
				sprite.draw(ctx);
			}
		}
	}
}

//dialogue box

function Dialogue(msgs, x, y, width){
	if (typeof msgs === "string") msgs = [msgs];
	this.msgs = msgs;
	this.i = 0;
	this.x = x;
	this.y = y;
	this.w = width;
	this.h = 0;
	this.lines = null;
}

Dialogue.prototype.advance = function(){
	this.i += 1;
	if (this.i >= this.msgs.length) return true;
	this.lines = null;
}

Dialogue.prototype.draw = function(ctx){
	ctx.font = "20pt Courier";
	if (this.lines === null){	//need to figure out word wrapping
		this.lines = [""];
		var words = this.msgs[this.i].match(/\S+/g), i = 0, j, metrics;
		for (j = 0; j < words.length; j++){
			metrics = ctx.measureText(this.lines[i]+words[j]);
			if (metrics.width+16 < this.w){
				this.lines[i] += words[j]+" ";
			}else{
				i += 1;
				this.lines.push(words[j]+" ");
			}
		}
		this.l = i+1;
		this.line_h = 26;
		this.h = 16+this.line_h*this.l;
	}
	
	ctx.beginPath();
	ctx.rect(this.x, this.y, this.w, this.h);
	ctx.fillStyle = "#000000";
	ctx.strokeStyle = "#090963";
	ctx.lineWidth = 4;
	ctx.fill();
	ctx.stroke();
	ctx.textAlign = "left";
	ctx.fillStyle = "#ffffff";
	for (var i = 0; i < this.l; i++){
		ctx.fillText(this.lines[i], this.x+8, this.y+(i+1)*this.line_h);
	}
}

//popup message

function Popup(msg, x, y){
	this.msg = msg;
	this.life = 2;	//how many seconds to live for
	this.x = x;
	this.y = y;
	this.w = null;
	this.h = 24;
}

Popup.prototype.draw = function(ctx){
	ctx.font = "16pt Courier";
	if (this.w === null) this.w = ctx.measureText(this.msg).width+16;
	ctx.beginPath();
	ctx.rect(this.x, this.y, this.w, this.h);
	ctx.fillStyle = "#000000";
	ctx.strokeStyle = "#090963";
	ctx.lineWidth = 4;
	ctx.fill();
	ctx.stroke();
	ctx.textAlign = "left";
	ctx.fillStyle = "#ffffff";
	ctx.fillText(this.msg, this.x+8, this.y+16);
}

//-------- game object

Game.prototype.init = function(){
	//make it pixellated
	if (typeof this.ctx.imageSmoothingEnabled === "boolean"){
		this.ctx.imageSmoothingEnabled = false;
	}else{
		this.ctx.mozImageSmoothingEnabled = false;
		this.ctx.webkitImageSmoothingEnabled = false;
		this.ctx.msImageSmoothingEnabled = false;
	}
	this.ctx.fillStyle = "#000000";
	this.ctx.fillRect(0, 0, this.width, this.height);
	
	//images
	this.imgs = {};
	var i, img, imgs = [
		"dude", "shadow", "teleporter", "nerd", "bush", "old-guy", "female", "skier", "crazy-guy",
		"floor0", "floor1", "floor2", "floor3", "floor4",
		"wall0", "wall1", "wall2", "wall3", "wall4"
	];
	for (i = 0; i < imgs.length; i++){
		this.progress++;
		img = new Image();
		img.addEventListener("load", this.run.bind(this), false);
		img.src = imgs[i]+".png";
		this.imgs[imgs[i]] = img;
	}
	Sprite.prototype._imgs = this.imgs;	//give all sprites access to the image list
	Room.prototype._imgs = this.imgs;
	
	//music
	this.bgm = new Audio("bgm-start.ogg"); 
	this.bgm_loop = new Audio("bgm-loop.ogg");
	this.bgm.addEventListener("ended", function(){this.bgm.currentTime = 0; this.bgm_loop.currentTime = 0; this.bgm_loop.play();}.bind(this), false);
	if (typeof this.bgm.loop === "boolean"){
		this.bgm_loop.loop = true;
	}else{
		this.bgm_loop.addEventListener("ended", function(){this.currentTime = 0; this.play();}, false);
	}
	
	//room data
	this.progress++;
	getJson("rooms.json", function(data){this.progress--, this.room_spec = data}.bind(this));
	this.room_num = 0;
	
	//objects
	this.player = null;
	this.room = null;
	this.rooms = null;
	this.dialogue = null;
	this.popup = null;
	
	//map stuff
	this.scale = 4;
	this.tile_size = 32*this.scale;
	this.grid_width = 12;
	this.grid_height = 8;
	this.buffer_right = this.width-200;
	this.buffer_left = 200;
	this.buffer_top = 160;
	this.buffer_bottom = this.height-160;
	
	//misc
	this.scene = 0;
	this.paused = false;
	this.just_unpaused = false;
	
	//events
	this.canvas.oncontextmenu = function(e) { // stop the right click menu
		e.preventDefault(); e.stopPropagation();
		return false;
	}
	document.addEventListener("keydown", this.handle_keyhit.bind(this), false);
	document.addEventListener("keyup", this.handle_keyup.bind(this), false);
	this.canvas.addEventListener("blur", function(){this.paused = true}.bind(this), true);
	this.canvas.addEventListener("focus", function(){
		if (this.paused === true){
			this.paused = false;
			this.just_unpaused = true;
			this.run();
		}
	}.bind(this), true);
	
	this.canvas.focus();
}

Game.prototype.reset = function(){
	var player, room, i;
	
	if (this.player === null){
		player = new Sprite("dude", 0, 0, this.scale, 36, 12, true);
		player.max_spd = 240;
		player.state_frames = {
			"up": [0, 1], "walk-up": [1, 9],
			"down": [9, 10], "walk-down": [10, 18],
			"left": [18, 19], "walk-left": [19, 27],
			"right": [27, 28], "walk-right": [28, 36]
		}
		player.grid_x = 6;
		player.grid_y = 5;
		player.inventory = ["Stick"];
		this.player = player;
	}else{
		player = this.player
	}
	player.x = (this.width-player.w)/2;
	player.y = (this.height-player.h)/2;
	player.spd = {x: 0, y: 0};
	player.state = "up";
	player.events = [];
	
	if (this.rooms === null){
		this.rooms = [];
		for (i = 0; i < this.room_spec.length; i++){
			room = new Room(this.room_spec[i], this.grid_width, this.grid_height, this.scale);
			room.sprites.push(player);
			this.rooms.push(room);
		}
	}
	
	room = this.rooms[this.room_num];
	room.x = player.x - player.grid_x*room.tile_size;
	if (room.x > 0){
		player.x -= room.x;
		room.x = 0
	}else if (room.x+room.full_width < this.width){
		player.x += this.width - (room.x+room.full_width);
		room.x = this.width - room.full_width;
	}
	room.y = player.bottom - (player.grid_y+1)*room.tile_size;
	if (room.y > 0){
		player.y -= room.y;
		room.y = 0;
	}else if (room.y+room.full_height < this.height){
		player.y += this.height - (room.y+room.full_height);
		room.y = this.height - room.full_height;
	}
	this.room = room;
	
	this.dialogue = null;
}

//-------- input

Game.prototype.handle_keyhit = function(e){
	var key = e.which || e.keyCode, handled = true;
	switch (key){
	case 32:	//space
		if (this.scene === 0){
			this.scene = 2;
			this.reset();
			this.run();
			this.bgm.play();
			var player = this.player, room = this.room, action = room.interaction[player.grid_x][player.grid_y];
			this.dialogue = new Dialogue(action.target.dialogue, 32, 32, this.width-64);
		}else if (this.scene === 1){
			var player = this.player, room = this.room, action = room.interaction[player.grid_x][player.grid_y], sprite, i, interact = false;
			if (action){
				if (action.type === "speak"){
					sprite = action.target;
					if (sprite.reaction){	//has yet to react - maybe it will happen now
						if (sprite.wants){
							i = player.inventory.indexOf(sprite.wants);
							if (i !== -1){
								player.inventory.splice(i, 1);
								interact = true;
							}
						}else{
							interact = true;
						}
						if (interact === true){
							if (sprite.response)	//says something else now
								sprite.dialogue = sprite.response;
							action = sprite.reaction;
							sprite.reaction = null;
							if (action.type === "activate"){	//activate a teleporter
								action = room.interaction[action.x][action.y]
								action.type = "teleport";
								action.sprite.state = "on";
								action.sprite.f = 1;
							}else if (action.type === "give"){	//give a new item
								player.inventory.push(action.item);
								this.popup = new Popup("You got the "+action.item+"!", 32, this.height-64);
							}
							if (sprite.nf > 1)	//hardcoded changing frame if there is one
								sprite.f = 1;
						}
					}
					this.dialogue = new Dialogue(sprite.dialogue, 32, 32, this.width-64);
					this.scene = 2;
				}else if (action.type === "teleport"){
					this.room_num = action.target.room;
					player.grid_x = action.target.x;
					player.grid_y = action.target.y;
					this.reset();
					this.popup = new Popup("Universe "+(this.room_num+1), 32, this.height-64);
				}
			}
		}else if (this.scene === 2){
			if (this.dialogue.advance() === true){
				this.dialogue = null;
				this.scene = 1;
			}
			this.run();
		}
		break;
	case 37: case 65:	//left / A
		if (this.scene === 1){	//playing
			if (this.player.events.indexOf("move-left") === -1)
				this.player.events.push("move-left");
		}
		break;
	case 38: case 87:	//up / W
		if (this.scene === 1){	//playing
			if (this.player.events.indexOf("move-up") === -1)
				this.player.events.push("move-up");
		}
		break;
	case 39: case 68:	//right / D
		if (this.scene === 1){	//playing
			if (this.player.events.indexOf("move-right") === -1)
				this.player.events.push("move-right");
		}
		break;
	case 40: case 83:	//down / S
		if (this.scene === 1){	//playing
			if (this.player.events.indexOf("move-down") === -1)
				this.player.events.push("move-down");
		}
		break;
	default:
		// console.log(key);
		handled = false;
	}
	if (handled === true){
		e.preventDefault(); e.stopPropagation();
	}
}

Game.prototype.handle_keyup = function(e){
	var key = e.which || e.keyCode;
	switch (key){
	case 37: case 65:	//left / A
		if (this.scene === 1){	//playing
			this.player.events.eradicate("move-left");
			this.player.state = "left";
		}
		break;
	case 38: case 87:	//up / W
		if (this.scene === 1){	//playing
			this.player.events.eradicate("move-up");
			this.player.state = "up";
		}
		break;
	case 39: case 68:	//right / D
		if (this.scene === 1){	//playing
			this.player.events.eradicate("move-right");
			this.player.state = "right";
		}
		break;
	case 40: case 83:	//down / S
		if (this.scene === 1){	//playing
			this.player.events.eradicate("move-down");
			this.player.state = "down";
		}
		break;
	}
}

//-------- update

Game.prototype.update = function(dt){
	var done = false;
	
	//handle un/pausing
	if (this.paused){
		done = true;
		if (this.bgm.currentTime > 0){
			this.bgm.pause();
		}else if (this.bgm_loop.currentTime > 0){
			this.bgm_loop.pause();
		}
	}else if (this.just_unpaused === true){
		if (this.bgm.currentTime > 0){
			this.bgm.play();
		}else if (this.bgm_loop.currentTime > 0){
			this.bgm_loop.play();
		}
	}
	
	if (this.scene === 0){	//splash screen
		done = true;
	}else if (this.scene === 1){	//playing
		var room = this.room, player = this.player, l = player.events.length, moving_h = false, moving_v = false, offset;
		player.spd.x = 0;
		player.spd.y = 0;
		if (l > 0){	//has events to handle
			for (i = l-1; i >= 0; i--){
				switch (player.events[i]){
				case "move-left":
					if (moving_h === false){
						player.spd.x = -player.max_spd;
						player.state = "walk-left";
						moving_h = true;
					}
					break;
				case "move-right":
					if (moving_h === false){
						player.spd.x = player.max_spd;
						player.state = "walk-right";
						moving_h = true;
					}
					break;
				case "move-up":
					if (moving_v === false){
						player.spd.y = -player.max_spd;
						player.state = "walk-up";
						moving_v = true;
					}
					break;
				case "move-down":
					if (moving_v === false){
						player.spd.y = player.max_spd;
						player.state = "walk-down";
						moving_v = true;
					}
					break;
				}
			}
		}
		if (moving_h === true && moving_v === true){	//reduce speed on diagonals
			player.spd.x /= Math.SQRT2;
			player.spd.y /= Math.SQRT2;
		}
		player.x += player.spd.x * dt;
		player.y += player.spd.y * dt;
		//collisions
		if (moving_h === true || moving_v === true){
			player.shadow.x += player.spd.x * dt;
			player.shadow.y += player.spd.y * dt;
			var rem_left = player.shadow.left-room.x, grid_left = Math.floor(rem_left/room.tile_size),
			rem_right = player.shadow.right-room.x, grid_right = Math.floor(rem_right/room.tile_size),
			rem_top = player.shadow.top-room.y, grid_top = Math.floor(rem_top/room.tile_size),
			rem_bottom = player.shadow.bottom-room.y, grid_bottom = Math.floor(rem_bottom/room.tile_size),
			coll = (room.occupied[grid_left][grid_top] << 3) | (room.occupied[grid_right][grid_top] << 2) | (room.occupied[grid_left][grid_bottom] << 1) | room.occupied[grid_right][grid_bottom],
			reaction_x = 0, reaction_y = 0;
			rem_left -= (grid_left+1)*room.tile_size;
			rem_right -= grid_right*room.tile_size;
			rem_top -= (grid_top+1)*room.tile_size;
			rem_bottom -= grid_bottom*room.tile_size;
			player.grid_x = Math.abs(rem_left) > Math.abs(rem_right) ? grid_left : grid_right;
			player.grid_y = Math.abs(rem_top) > Math.abs(rem_bottom) ? grid_top : grid_bottom;
			if (coll > 0){
				if (player.spd.x < 0){
					if (coll & 0b1010) reaction_x = rem_left;
				}else if (player.spd.x > 0){
					if (coll & 0b0101) reaction_x = rem_right;
				}
				if (player.spd.y < 0){
					if (coll & 0b1100) reaction_y = rem_top;
				}else if (player.spd.y > 0){
					if (coll & 0b0011) reaction_y = rem_bottom;
				}
				if ((coll & 0b1001) === 0b1001 || (coll & 0b0110) === 0b0110){	//definitely need both
					player.x -= reaction_x;
					player.y -= reaction_y;
				}else{
					if (coll === 0b1010 || coll === 0b0101)	//horizontal only
						player.x -= reaction_x;
					else if (coll === 0b1100 || coll === 0b0011)	//vertical only
						player.y -= reaction_y;
					else{	//whichever is smallest (and nonzero)
						if (reaction_x !== 0 && reaction_y !== 0){
							if (Math.abs(reaction_x) < Math.abs(reaction_y))
								reaction_y = 0;
							else
								reaction_x = 0;
						}
						player.x -= reaction_x;
						player.y -= reaction_y;
					}
				}
			}
		}
		
		//scroll the room around if necessary
		if (player.spd.x > 0){
			offset = Math.min(player.right-this.buffer_right, room.full_width-this.width+room.x);
			if (offset > 0){
				room.x -= offset;
				player.x -= offset;
			}
		}else if (player.spd.x < 0){
			offset = Math.max(player.left-this.buffer_left, room.x);
			if (offset < 0){
				room.x -= offset;
				player.x -= offset;
			}
		}
		if (player.spd.y > 0){
			offset = Math.min(player.bottom-this.buffer_bottom, room.full_height-this.height+room.y);
			if (offset > 0){
				room.y -= offset;
				player.y -= offset;
			}
		}else if (player.spd.y < 0){
			offset = Math.max(player.top-this.buffer_top, room.y);
			if (offset < 0){
				room.y -= offset;
				player.y -= offset;
			}
		}
		player.animate(dt);
		
		if (this.popup !== null){
			this.popup.life -= dt;
			if (this.popup.life < 0) this.popup = null;
		}
	}else if (this.scene === 2){
		done = true;
	}
	this.just_unpaused = false;
	
	this.render();
	return done;
}

Game.prototype.render = function(){
	var ctx = this.ctx;
	
	if (this.scene === 0){	//title screen
		ctx.textAlign = "center";
		ctx.fillStyle = "#ffffff";
		ctx.font = "14pt Courier";
		ctx.fillText("Welcome to", this.width/2, this.height/4);
		ctx.font = "Bold 40pt Courier";
		ctx.fillText("Inter-universal", this.width/2, this.height/3);
		ctx.fillText("Treasure Hunter", this.width/2, this.height/3+64);
		ctx.font = "16pt Courier";
		ctx.fillText("Press [SPACE] to begin", this.width/2, this.height*3/4);
	}else if (this.scene === 1 || this.scene === 2){	//playing/speaking to someone
		this.room.draw(ctx);
		if (this.dialogue !== null) this.dialogue.draw(ctx);
		if (this.scene === 1){	//only draw popup in playing mode
			if (this.popup !== null) this.popup.draw(ctx);
		}
	}	
	//pause
	if (this.scene > 0 && this.paused === true){
		ctx.fillStyle = "rgba(0, 0, 0, 0.5)";
		ctx.fillRect(0, 0, this.width, this.height);
		ctx.textAlign = "center";
		ctx.fillStyle = "#ffffff";
		ctx.font = "Bold 40pt Courier";
		ctx.fillText("Paused...", 400, 300);
	}
}

//--------

window.onload = function(){
	game = new Game();
}
